import .syntax
import .desugar_wf

/-
  Our goal is to "desugar" SugarImp programs into
  equivalent Imp programs.  Concretely, we want our
  translation to satisfy the following property:

<<
    If SugarImp program P produces output FOO and
    returns result RES when translated to Python
    and run on some input, then (desugar P) will
    also produce output FOO and return result RES
    when translated to Python and run on the same
    input.
>>

  This is primarily tricky because we have to "split"
  some SugarImp constructs into multiple, simpler
  Imp constructs.  For example, we have to split
  SugarImp expressions containining calls into a
  sequence of Imp statements:

<<
    The SugarImp expression:

      foo(1, 2, bar("hi"))

    should translate to something (conceptually) like:

      temp0 = bar("hi")
      temp1 = foo(1, 2, temp0)

    In practice, our translation will introduce
    many more temporary variables to make the
    translation code simpler.
>>

  Managing these "fresh" temporary names is a little
  tedious, but we must do it if we want our
  translation to produce equivalent output!
-/

def char.next (c : char) : char :=
  char.of_nat $ (char.to_nat c) + 1

def bump_temp_char (c : char) : string :=
  if '0' <= c ∧ c <= '8'
  then string.str "" c.next
  else "90"

def bump_temp_aux : list char → list char
| [] := "temp_0".to_list
| (c :: []) := (bump_temp_char c).to_list -- fix me
| (c :: s) := c :: bump_temp_aux s

def bump_temp (s : string) : string :=
  (bump_temp_aux s.to_list).as_string

/-
   As a concrete example, we have provided a function
   that desugars expressions. Every time we desugar an
   SugarImp expression [E] we get three things:
   (1) the Imp stmt that corresponds to running [E],
   (2) the name where the result of running [E] is stored, and
   (3) the name of the next "fresh" temporary we can use.
   Given these pieces, [desugar_expr] just desugars any
   subexpressions, then adds a new Imp stmt that puts
   their results together in a way that is equivalent
   to the original SugarImp expression.
-/

def mk_seq : list imp.stmt → imp.stmt
| [] := imp.stmt.nop
| (s :: ss) := imp.stmt.seq s (mk_seq ss)

-- PROBLEM 5 : 10 points

/- we are defining mutually recursive functions here -/
mutual def desugar_expr, desugar_call_args

/- the first is desugar_expr, which does most of the heavy lifting -/
with desugar_expr : sugar.expr
                  → string
                  → (imp.stmt × string × string)
| (sugar.expr.val v) t :=
    -- TODO complete this case (about 2 lines)
    (imp.stmt.nop, t, t)

| (sugar.expr.var x) t :=
    let
      x' := "orig_" ++ x,
      t1 := bump_temp t
    in
      ( imp.stmt.set t (imp.expr.var x')
      , t
      , t1
      )
| (sugar.expr.op1 op e) t :=
    let
      (s1, res1, t1) := desugar_expr e t,
      t2 := bump_temp t1
    in
      ( imp.stmt.seq s1
          (imp.stmt.set t1
            (imp.expr.op1 op (imp.expr.var res1)))
      , t1
      , t2
      )
| (sugar.expr.op2 op e1 e2) t :=
    let
      (s1, res1, t1) := desugar_expr e1 t,
      (s2, res2, t2) := desugar_expr e2 t1,
      t3 := bump_temp t2
    in
      ( mk_seq [s1, s2,
          imp.stmt.set t2
            (imp.expr.op2 op (imp.expr.var res1) (imp.expr.var res2))]
      , t2
      , t3
      )

| (sugar.expr.length e) t :=
    -- TODO complete this case
    -- HINT: this is very similar to the op1 case above!
    (imp.stmt.nop, t, t)

| (sugar.expr.index e1 e2) t :=
    -- TODO complete this case
    -- HINT: this is very similar to the op2 case above!
    (imp.stmt.nop, t, t)

| (sugar.expr.alloc e1 e2) t :=
    -- TODO complete this case
    -- HINT: this is pretty similar to the op2 case above!
    (imp.stmt.nop, t, t)

  /-
     This case is trickier because we need to plumb
     the "next temporary" through all the desugars
     of the argument expressions.  It is not conceptually
     different from the ealier cases, but does require
     using the other part of our mutually recursive
     definition, which essentially is just doing fold
     over lists.

     NOTE: we must respect Python's order of evaluating
     function arguments, which is left-to-right.
  -/
| (sugar.expr.call f es) t :=
  let
    (s, rs, nt) := desugar_call_args es (imp.stmt.nop, [], t),
    t1 := bump_temp nt
  in
    ( imp.stmt.seq s (imp.stmt.call nt f rs.reverse)
    , nt
    , t1
    )

/- Down here we have the second part of our mutually recursive
   function definitions.  desugar_call_args is essentially just
   doing a fold over a list of expressions for us. -/
with desugar_call_args : list sugar.expr
                       → (imp.stmt × list imp.expr × string)
                       → (imp.stmt × list imp.expr × string)
| []  ⟨ s, rs, nt ⟩ :=
    ⟨ s, rs, nt ⟩
| (e :: es) ⟨ s, rs, nt ⟩ :=
    let
      res := desugar_expr e nt,
      t2 := bump_temp (res.snd.snd)
    in
      desugar_call_args es
        ( imp.stmt.seq s res.fst
        , (imp.expr.var res.snd.fst) :: rs
        , t2
        )

/- do not look in the file where this tactic is defined,
   it is only suited for those who have been initiated,
   brought into the world of deep wizardry,
   from whence few return -/
using_well_founded desugar_wf.tactics

/--
  Convenience helper that starts desugaring
  with the right initial temp var, and throws
  away final "next temp" info.
-/
def desugar_expr' (e : sugar.expr) : (imp.stmt × string) :=
  let
    (s, res, _) := desugar_expr e "temp_0"
  in
    (s, res)

/-
  The tricky bit of desugaring statements are the
  early returns.  We will adopt the following strategy:
  first, we assume two variable names "ret_flg" and
  "ret_res" are available which indicate when a return has happened
  what its result is.  Then for each statement we
  translate, we wrap it in a branch that tests to see
  whether a return has happened and only executes the
  translated code if no return has happened.

  For example this simple factorial function (in the concrete syntax):
<<
    def fact(n) {
      if(n < 2) {
        return 1;
      } else {
        return n * fact(n - 1);
      }
    }
>>
  will desugar to something like (in the concrete syntax):
<<
    def fact(orig_n) {
      ret_flg = False;
      ret_res = 0;
      if (not ret_flg) {
        temp_0 = orig_n;
        temp_1 = 2;
        temp_2 = (temp_0 < temp_1);
        if temp_2 {
          if (not ret_flg) {
            temp_0 = 1;
            ret_flg = True;
            ret_res = temp_0;
          } else {
            nop;
          }
        } else {
          if (not ret_flg) {
            temp_0 = orig_n;
            nop;
            temp_1 = orig_n;
            temp_2 = 1;
            temp_3 = (temp_1 - temp_2);
            temp_5 = fact(temp_3);
            temp_6 = (temp_0 * temp_5);
            ret_flg = True;
            ret_res = temp_6;
          } else {
            nop;
          }
        }
      } else {
        nop;
      }
      return ret_res;
    }
>>

  We know that we can safely use the variable names
  "ret_flg" and "ret_res" because we have been careful
  to prepend "orig_" to all the variable names from
  the input SugarImp program.
-/

-- PROBLEM 6 : 10 points
def desugar_stmt_inner : sugar.stmt → imp.stmt
| (sugar.stmt.expr e) :=
    -- TODO complete this case
    -- HINT: this is pretty similar to the set case below!
    imp.stmt.nop

| (sugar.stmt.set x e) :=
    let
      x' := "orig_" ++ x,
      (s', res) := desugar_expr' e
    in
      imp.stmt.seq s'
        (imp.stmt.set x' (imp.expr.var res))
| (sugar.stmt.write e1 e2 e3) :=
    let
      (s1, res1, t1) := desugar_expr e1 "temp_0",
      (s2, res2, t2) := desugar_expr e2 t1,
      (s3, res3, t3) := desugar_expr e3 t2
    in
       mk_seq [s1, s2, s3,
         imp.stmt.write res1 (imp.expr.var res2) (imp.expr.var res3)]
| (sugar.stmt.ifthen e ss1) :=
    let
      (s0, res0) := desugar_expr' e,
      s1 := desugar_stmt_inner ss1
    in
      mk_seq [s0,
        imp.stmt.ifelse (imp.expr.var res0) s1 imp.stmt.nop]

| (sugar.stmt.ifelse e ss1 ss2) :=
    -- TODO complete this case
    -- HINT: this is pretty similar to the ifthen case above!
    imp.stmt.nop

| (sugar.stmt.while e ss1) :=
    let
      (s0, res0) := desugar_expr' e,
      s1 := desugar_stmt_inner ss1,
      guard := imp.expr.op2 op2.conj
                 (imp.expr.var res0)
                 (imp.expr.op1 op1.not (imp.expr.var "ret_flg"))
    in
      mk_seq [s0,
        imp.stmt.while guard (imp.stmt.seq s1 s0)]

| (sugar.stmt.seq ss1 ss2) :=
    -- TODO complete this case
    -- HINT: this is pretty similar to the ifelse case you did above!
    imp.stmt.nop

| (sugar.stmt.ret e) :=
    let
      (s', res) := desugar_expr' e
    in
      mk_seq [s',
        imp.stmt.set "ret_flg" $ imp.expr.val (val.bool tt),
        imp.stmt.set "ret_res" $ imp.expr.var res]

def desugar_stmt (s : sugar.stmt) : imp.stmt :=
  imp.stmt.ifelse
    (imp.expr.op1 op1.not (imp.expr.var "ret_flg"))
    (desugar_stmt_inner s)
    imp.stmt.nop.

def desugar_func : sugar.func → imp.func
| (sugar.func.mk name args body) :=
  let
    args' := args.map (fun x, "orig_" ++ x),
    body' := mk_seq [
      imp.stmt.set "ret_flg" (imp.expr.val (val.bool false)),
      imp.stmt.set "ret_res" (imp.expr.val (val.int 0)),
      (desugar_stmt body)
    ]
  in
    imp.func.mk name args' body' (imp.expr.var "ret_res").

def desugar_prog : sugar.prog → imp.prog
| (sugar.prog.mk funcs main) :=
  let
    funcs' := funcs.map desugar_func
  in
    imp.prog.mk funcs' main
