import ..lexer
import ..parser.core
import ..syntax
import ..parser.expr

namespace sugar
namespace parser

open parser
open lexer.tok

@[reducible] def t := parser lexer.tok

def is_ident : lexer.tok → bool
| (lexer.tok.ident _) := true
| _ := false

def is_int : lexer.tok → bool
| (lexer.tok.int_lit _) := true
| _ := false

def is_str : lexer.tok → bool
| (lexer.tok.str_lit _) := true
| _ := false

def ident : t string :=
do (lexer.tok.ident i) ← sat (λ t, is_ident t = tt) | parser.fail "not ident",
   return i

def int_lit : t int :=
do tk ← sat (λ t, is_int t = tt),
   match tk with
   | (lexer.tok.int_lit i) := return i
   | _ := parser.fail "expected integer literal"
   end

def str_lit : t string :=
do (lexer.tok.str_lit s) ← sat (λ t, is_str t = tt) | parser.fail "expected string",
   return s

def bracket {elem α : Type} : parser elem unit → parser elem unit → parser elem α → parser elem α :=
fun o c p, do o, r ← p, c, pure r

def curlies {α : Type} : t α → t α :=
bracket (sat (= lcurly) *> pure ()) (sat (= rcurly) *> pure ())

def parens {α : Type} (p : t α) : t α :=
bracket (sat (= lparen) *> pure ()) (sat (= rparen) *> pure ()) p

def squares {α : Type} (p : t α) : t α :=
sat (= lsquare) *> p <* sat (= rsquare)

def params : t (list string) :=
parens (sep_by (sat (= comma) *> pure ()) ident)

def binary_op (tk : lexer.tok) (o : op2) (assoc : parser.expr.assoc) : expr.operator lexer.tok expr :=
parser.expr.operator.inf (do sat (= tk), return (expr.op2 o)) assoc

def unary_op (tk : lexer.tok) (o : op1) : expr.operator lexer.tok expr :=
parser.expr.operator.pre (do sat (= tk), return (expr.op1 o))

def assign_to {res l r : Type} (lhs : t l) (rhs : t r) (cont : l → r → t res) : t res :=
    do l ← lhs,
       sat (= assign),
       r ← rhs,
       sat (= semi),
       cont l r

section expression

parameter Exp : t expr

def pcall : t expr :=
decorate_error "call expression" $
  do fn ← ident,
     args ← parens $ sep_by (sat (= comma) *> pure ()) Exp,
     pure $ sugar.expr.call fn args

def alloc : t expr :=
  do sat (= alloc),
     parens (do
        e1 ← Exp,
        sat (= lexer.tok.comma),
        e2 ← Exp,
        pure $ expr.alloc e1 e2)

def atomic_expr : t expr :=
  parens Exp <|>
  sat (= lexer.tok.true) *> pure (expr.val (val.bool tt)) <|>
  sat (= lexer.tok.false) *> pure (expr.val (val.bool ff)) <|>
  expr.val <$> val.int <$> int_lit <|>
  expr.val <$> val.str <$> str_lit <|>
  alloc <|>
  (do sat (= lexer.tok.len), e ← parens Exp, return (expr.length e)) <|>
  (try (do e ← expr.var <$> ident,
   i ← squares Exp,
   pure (expr.index e i))) <|>
  try pcall <|>
  ident >>= λ i, pure (expr.var i)

def ops : expr.operator_table lexer.tok expr := [
   [unary_op lexer.tok.sub op1.neg, unary_op lexer.tok.not op1.not],
   [binary_op lexer.tok.div op2.div expr.assoc.left, binary_op lexer.tok.mul op2.mul expr.assoc.left, binary_op lexer.tok.mod op2.mod expr.assoc.left],
   [binary_op lexer.tok.add op2.add expr.assoc.left, binary_op lexer.tok.sub op2.sub expr.assoc.left],
   [binary_op lexer.tok.le op2.le expr.assoc.none, binary_op lexer.tok.lt op2.lt expr.assoc.none],
   [binary_op lexer.tok.eq op2.eq expr.assoc.right],
   [binary_op lexer.tok.disj op2.disj expr.assoc.left, binary_op lexer.tok.conj op2.conj expr.assoc.none]
]

def binary_ops := expr.mk_expr_parser ops atomic_expr

end expression

def expression : t expr :=
fix (λ exp, binary_ops exp)

def index_of : t (expr × expr) :=
  do i ← ident,
     sat (= lsquare),
     e ← expression,
     sat (= rsquare),
     return (expr.var i, e)

def if_st (st : t stmt) : t stmt :=
    do sat (= if_t),
       guard ← parens expression,
       true_blk ← curlies st,
       (do sat (= else_t),
            false_blk ← curlies st,
            return $ stmt.ifelse guard true_blk false_blk) <|>
       (do return $ stmt.ifthen guard true_blk)

def statement (st : t stmt) : t stmt :=
    stmt.ret <$> (sat (= ret_t) *> expression <* sat (= semi)) <|>
    try (assign_to ident expression (λ i e, pure $ stmt.set i e)) <|>
    try (assign_to index_of expression (λ ⟨a, i⟩ e, pure $ stmt.write a i e)) <|>
    stmt.expr <$> expression <* sat (= semi) <|>
    if_st st <|>
    sat (= while) *> (stmt.while <$> parens expression <*> curlies st)

def block : t stmt :=
parser.fix (fun st,
do ss ← many1 $ statement st,
   match ss with
   | [] := parser.failure
   | (s :: ss) := pure $ ss.foldl stmt.seq s
   end)

def fn : t func :=
do sat (= lexer.tok.def_t),
   i ← ident,
   ps ← params,
   sat (= lcurly),
   body ← block,
   sat (= rcurly),
   return (func.mk i ps body)

def file : t prog :=
do fns ← many fn,
   i ← ident,
   return $ prog.mk fns i

def pcompose {a b c d : Type} [has_to_string c] (f : b → buffer c) (p1 : parser a b) (p2 : parser c d) : parser a d :=
do res1 ← p1,
   match p2.run (f res1) with
   | sum.inl msg := parser.fail msg
   | sum.inr v := pure v
   end

def parse_buffer {α : Type} (s : char_buffer) (p : t α) : sum string α :=
let tokens := lexer.tokenize s
in trace (to_string tokens) (p.run tokens.to_buffer)

def program : parser char prog :=
pcompose (fun (l : list lexer.tok), l.to_buffer) lexer.tokenize' (file <* sat (= lexer.tok.eof))

def parse_string {α : Type} (s : string) (p : t α) : sum string α :=
 let tokens := lexer.tokenize s.to_list.to_buffer
 in trace (to_string tokens) (p.run tokens.to_buffer)

def parse_eof (s : char_buffer) : sum string func :=
 let tokens := lexer.tokenize s
 in ((fn <* sat (=lexer.tok.eof)).run tokens.to_buffer)

end parser
end sugar
