/-

  In this file we define two programming languages: SugarImp and Imp.

  SugarImp has many of the core features of Python, including functions
  and arrays.  Imp also has these same features, but is a simplified language:
  expressions are always pure, every function has exactly one return at the
  very end, and array writes must be through a local variable.

  Our goal in this file is to provide a translation from SugarImp to Imp.
  Most of the scaffolding has already been provided.  Your goal is to fill
  in the TODOs we've left for you below.

  Please note that this is a new assignment, so there may be a few kinks.
  Please let us know if you notice anything weird or buggy and we'll try
  to get a fix pushed pronto!

  We have also provided shim code in the ml/ directory so you can translate
  SugarImp and Imp programs to Python and run them.

  To build the shim, make sure you are one directory above where this
  file is stored and run:
<<
    $ leanpkg build
>>
  Then, to run SugarImp program "foo.simp", first translate it to Python:
<<
    $ lean --run src/main.lean --Spython foo.simp > foo.py
>>
  and run it with Python:
<<
    $ python foo.py
>>
  Similarly, you can translate and run Imp program "bar.imp" by:
<<
    $ lean --run src/main.lean --Ipython bar.imp > bar.py
    $ python bar.py
>>
  Note that SugarImp and Imp programs have library functions you
  can call to perform I/O including:
<<
    print_val(VAL)
    read_bool()
    read_int()
    read_string()
>>
  For more examples, please see the test/ directory.

  Finally, you can test your translation by running:
<<
    $ make test
>>
  in the directory one above where this file is stored.

-/




-- Define common types.

/- In a source program, a value will correspond to a constant. -/
inductive val : Type
| bool : bool → val
| int  : int → val
| str  : string → val.

/- Both SugarImp and Imp have the same unary
   and binary operators, defined below in types
   `op1` and `op2` respectively. -/
inductive op1 : Type
| neg
| not

inductive op2 : Type
| add
| sub
| mul
| div
| mod
| eq
| lt
| le
| conj
| disj

section string_instances

open val

def val.to_string : val → string
 | (bool tt) := "True"
 | (bool ff) := "False"
 | (int i) := to_string i
 | (str s) := repr s

instance val.has_to_string : has_to_string val :=
⟨ val.to_string ⟩

open op1

def op1.to_string : op1 → string
| neg := "-"
| not := "not"

instance op1.has_to_string : has_to_string op1 :=
⟨ op1.to_string ⟩

open op2
def op2.to_string : op2 → string
| add := "+"
| sub := "-"
| mul := "*"
| div := "/"
| mod := "%"
| eq := "=="
| lt := "<"
| le := "<="
| conj := "and"
| disj := "or"

instance op2.has_to_string : has_to_string op2 :=
⟨ op2.to_string ⟩

def val.repr : val → string
 | (bool tt) := "(val.bool tt)"
 | (bool ff) := "(val.bool ff)"
 | (int i) := "(int " ++ repr i ++ ")"
 | (str s) := "(str " ++ repr s ++ ")"

instance val.has_repr : has_repr val :=
⟨ val.repr ⟩

open op1

def op1.repr : op1 → string
| neg := "neg"
| not := "not"

instance op1.has_repr : has_repr op1 :=
⟨ op1.repr ⟩

open op2

def op2.repr : op2 → string
| add := "add"
| sub := "sub"
| mul := "mul"
| div := "div"
| mod := "mod"
| eq := "eq"
| lt := "lt"
| le := "le"
| conj := "conj"
| disj := "disj"

instance op2.has_repr : has_repr op2 :=
⟨ op2.repr ⟩

end string_instances

namespace sugar
/-
  Now we define expressions in SugarImp.
  Unlike in Imp (which we define further below),
  SugarImp expressions include function calls
  and array allocation and thus can have side
  effects.
-/
inductive expr : Type

-- constants, e.g., `42`
| val : val -> expr

-- variables, like `x`
| var : string -> expr

-- unary operation, like `not e`
| op1 : op1 -> expr -> expr

-- binary operation, like `e1 + e2`
| op2 : op2 -> expr -> expr -> expr

-- array lengths, like `len(e)`
| length : expr -> expr

-- array indexing, like `e1[e2]`
| index : expr -> expr -> expr

-- function calls, like [f(e1, ..., eN)]
| call : string -> list expr -> expr

/-

 array allocation, in Python [e1 * [e2]], e.g.

>>> 5 * ['hi']
['hi', 'hi', 'hi', 'hi', 'hi']

-/
| alloc : expr -> expr -> expr.

/-

  Next we define statements in SugarImp.
  Unlike in Imp (which we define further below),
  SugarImp statements include expressions,
  "early return" (before the end of a function),
  and "multidimensional" array writes, like

<<
    x[1][2][3] = "blue"
>>

-/

inductive stmt : Type

-- expression statement, like [f(e1, ..., eN)]
| expr : expr -> stmt

-- assigning to a variable, like [x = 3]
| set : string -> expr -> stmt

-- writing to an array, like [a[0] = 17]
| write : expr -> expr -> expr -> stmt

/- conditional branch, like:

   if e:
    s1

-/
| ifthen : expr -> stmt -> stmt

/- conditional branch with else, like:

  if e:
    s1
  else:
    s2

-/
| ifelse : expr -> stmt -> stmt -> stmt

/-
  while loop, like

  while e:
    s

-/
| while : expr -> stmt -> stmt

/-
  statement sequencing, like

  s1
  s2

-/
| seq : stmt -> stmt -> stmt

-- return, like [return 5]
| ret : expr -> stmt.


/- functions, like

  def f(x1, ..., xN):
    s

-/
inductive func : Type
| mk : string -> list string -> stmt -> func.

/-

  Programs which are just a list of functions
  and the name of the "main" function.

  NOTE: we will assume the main function takes no arguments.

-/
inductive prog : Type
| mk : list func -> string -> prog

end sugar



/-

  Imp programs are simplified SugarImp programs.
  In particular, expressions are pure (never have side effects),
  heap writes are directly through a local variable
  (no direct "multidimensional" array writes), and functions
  only (and always) return at the end.  We also have
  an additional kind of statement, [Snop], which does
  nothing.

-/

namespace imp

inductive expr : Type
| val : val -> expr
| var : string -> expr
| op1 : op1 -> expr -> expr
| op2 : op2 -> expr -> expr -> expr
| length : expr -> expr
| index : expr -> expr -> expr.

inductive stmt : Type
| nop : stmt
| set : string -> expr -> stmt
| alloc : string -> expr -> expr -> stmt
| write : string -> expr -> expr -> stmt
| call : string -> string -> list expr -> stmt
| ifelse : expr -> stmt -> stmt -> stmt
| while : expr -> stmt -> stmt
| seq : stmt -> stmt -> stmt.

/- functions that always end with return, like

  def f(x1, ..., xN):
    s
    return e

-/
inductive func : Type
| mk : string -> list string -> stmt -> expr -> func

inductive prog : Type
| mk : list func -> string -> prog.

end imp
